use btlib::{
    self, bterr,
    crypto::{file_cred_store::FileCredStore, tpm::TpmCredStore, CredStore},
    Result,
};
use config::{
    builder::{AsyncState, DefaultState},
    ConfigBuilder, Environment,
};
use serde::{de::Visitor, Deserialize, Deserializer};
use std::{path::PathBuf, result::Result as StdResult, str::FromStr};

/// Attempts to unwrap the [Option] field in `$config` called `$name`. This macro internal
/// uses the `?` operator, and can only be used in method which return a [Result] which has an
/// error type that a [btlib::Error] can be converted to.
#[macro_export]
macro_rules! get_setting {
    ($config:expr, $name:ident) => {
        $config
            .$name
            .ok_or_else(|| bterr!(concat!("setting '", stringify!($name), "' was not set")))?
    };
}

/// Attempts to unwrap all of the [Option] fields with the given names in the configuration struct
/// and returns them all as a tuple. The same consideration which apply to [get_setting!] apply to
/// this macro.
#[macro_export]
macro_rules! get_settings {
    ($config:expr$(, $names:ident)+) => {
        ($(get_setting!($config, $names),)+)
    };
}

fn environment_source() -> Environment {
    Environment::with_prefix("BT").separator("_")
}

/// An extension to [ConfigBuilder] which allows it to be configured using Blocktree's default
/// settings.
pub trait ConfigBuilderExt {
    /// Configure this [ConfigBuilder] with Blocktree's default settings.
    fn btconfig(self) -> Self;
}

impl ConfigBuilderExt for ConfigBuilder<DefaultState> {
    fn btconfig(self) -> Self {
        self.add_source(environment_source())
    }
}

impl ConfigBuilderExt for ConfigBuilder<AsyncState> {
    fn btconfig(self) -> Self {
        self.add_source(environment_source())
    }
}

/// A struct which assists with very simple parsing of a configuration string.
struct StrParser<'a>(&'a str);

impl<'a> StrParser<'a> {
    /// Assert that the given token occurs at the current position in the input `str`.
    fn assert(&mut self, token: &str) -> btlib::Result<()> {
        if self.0.starts_with(token) {
            self.0 = &self.0[token.len()..];
            Ok(())
        } else {
            Err(bterr!("string does not start with {token}"))
        }
    }

    /// Consumes all characters up to `stop`, and returns them. Note that the returned `str` does
    /// not contain `stop`, and the current position in the `str` is set to the next character after
    /// `stop`. If `stop` is not encountered, then the remaining `str` is returned.
    fn consume_up_to(&mut self, stop: char) -> &str {
        let mut count = 0;
        for c in self.0.chars() {
            if c == stop {
                break;
            }
            count += 1;
        }
        let output = &self.0[..count];
        let new_start = self.0.len().min(count + 1);
        self.0 = &self.0[new_start..];
        output
    }

    /// Returns the remaining unconsumed part of the `str`.
    fn remaining(&self) -> &str {
        self.0
    }
}

#[derive(Debug, Clone)]
/// Configuration which specifies how to construct a [btlib::crypto::CredStore].
pub enum CredStoreCfg {
    /// Specifies that a [btlib::crypto::file_cred_store::FileCredStore] should be used.
    File {
        /// The path where the credential store is located.
        path: PathBuf,
    },
    /// Specifies that a [btlib::crypto::tpm::TpmCredStore] should be used.
    Tpm {
        /// The state path to use for the TPM credential store.
        tpm_state_path: PathBuf,
        /// The configuration string to pass to `tss-esapi`.
        tabrmd: String,
    },
}

pub trait CredStoreConsumer {
    type Output;
    fn consume<C: CredStore>(self, cred_store: C) -> Self::Output;
}

impl CredStoreCfg {
    /// Calls `consumer` with the [CredStore] created based on the configuration data in `self`.
    pub fn consume<F: CredStoreConsumer>(self, consumer: F) -> Result<F::Output> {
        match self {
            CredStoreCfg::File { path } => {
                let store = FileCredStore::new(path)?;
                Ok(consumer.consume(store))
            }
            CredStoreCfg::Tpm {
                tpm_state_path,
                tabrmd,
            } => {
                let store = TpmCredStore::from_tabrmd(&tabrmd, tpm_state_path)?;
                Ok(consumer.consume(store))
            }
        }
    }
}

impl FromStr for CredStoreCfg {
    type Err = btlib::Error;
    fn from_str(value: &str) -> StdResult<Self, Self::Err> {
        let mut parser = StrParser(value);
        parser.assert("kind=")?;
        let kind = parser.consume_up_to(',');
        match kind {
            "file" => {
                parser.assert("path=")?;
                let path = PathBuf::from(parser.remaining());
                Ok(CredStoreCfg::File { path })
            }
            "tpm" => {
                parser.assert("tpm_state_path=")?;
                let tpm_state_path = PathBuf::from(parser.consume_up_to(','));
                let tabrmd = parser.remaining().to_string();
                Ok(CredStoreCfg::Tpm {
                    tpm_state_path,
                    tabrmd,
                })
            }
            _ => Err(bterr!(
                "unrecognized CredStore kind (expected 'file' or 'tpm'): {kind}"
            )),
        }
    }
}

impl<'de> Deserialize<'de> for CredStoreCfg {
    fn deserialize<D: Deserializer<'de>>(deserializer: D) -> StdResult<Self, D::Error> {
        struct StructVisitor;

        impl<'a> Visitor<'a> for StructVisitor {
            type Value = CredStoreCfg;

            fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                write!(formatter, "cred store configuration string")
            }

            fn visit_str<E: serde::de::Error>(self, v: &str) -> StdResult<Self::Value, E> {
                CredStoreCfg::from_str(v).map_err(serde::de::Error::custom)
            }
        }

        deserializer.deserialize_str(StructVisitor)
    }
}

#[cfg(test)]
mod tests {
    use serde::de::value::StrDeserializer;

    use super::*;

    #[test]
    fn deserialize_file_cred_store_config() {
        const PATH: &str = "./state/cred_store";
        let input = format!("kind=file,path={PATH}");
        let deserializer = StrDeserializer::<'_, serde::de::value::Error>::new(input.as_str());

        let actual = CredStoreCfg::deserialize(deserializer).unwrap();

        let passed = if let CredStoreCfg::File { path } = actual {
            path.to_str().unwrap() == PATH
        } else {
            false
        };
        assert!(passed);
    }

    #[test]
    fn deserialize_tpm_cred_store_config() {
        const STATE_PATH: &str = "./state/cred_store";
        const TABRMD: &str = "bus_type=session";
        let input = format!("kind=tpm,tpm_state_path={STATE_PATH},{TABRMD}");
        let deserializer = StrDeserializer::<'_, serde::de::value::Error>::new(input.as_str());

        let actual = CredStoreCfg::deserialize(deserializer).unwrap();

        let passed = if let CredStoreCfg::Tpm {
            tpm_state_path,
            tabrmd,
        } = actual
        {
            tpm_state_path.to_str().unwrap() == STATE_PATH && &tabrmd == TABRMD
        } else {
            false
        };
        assert!(passed);
    }
}
